"""Internal implementation details.

.. warning::

    This package is intended only for internal use within Accretion.
    This documentation is provided for information purposes only.
    No guarantee is provided on the modules and APIs described in here remaining consistent.
    Directly reference at your own risk.
"""
__all__ = ("__version__",)

__version__ = "0.0.1b0"



"""Create a venv, build requirements into it, and package that into a zip."""
import io
import logging
import os
from zipfile import ZipFile

__all__ = ("build_zip",)
_LOGGER = logging.getLogger(__name__)


def _file_filter(filename: str) -> bool:  # pylint: disable=unused-argument
    """Determine whether this file should be included in the zip file.

    :param str filename: Filename
    :rtype: bool
    """
    # For now, leave everything in.
    return True


#    Leaving this here as an example of what this function might do later.
#    if filename.endswith(".pyc"):
#        return False

#    if f"{os.path.sep}__pycache__{os.path.sep}" in filename:
#        return False

#    return True


def build_zip(build_dir: str) -> io.BytesIO:
    """Build a Lambda Layer zip from a given directory.

    :param str build_dir: Directory to pack into Layer zip
    :returns: Binary file-like of zip
    :rtype: io.BytesIO
    """
    buffer = io.BytesIO()
    with ZipFile(buffer, mode="w") as zipper:
        for root, _dirs, files in os.walk(build_dir):
            for filename in files:
                if not _file_filter(filename):
                    continue

                filepath = os.path.join(root, filename)
                zipper.write(filename=filepath, arcname=f"python/{filepath[len(build_dir) + 1:]}")

    _LOGGER.debug("Zip file size: %s bytes", buffer.tell())
    buffer.seek(0)
    return buffer



"""Internal Accretion exceptions."""


class AccretionError(Exception):
    """Parent Accretion exception."""


class ZipBuilderError(AccretionError):
    """Raised when an error is encountered building a zip file."""


class ExecutionError(AccretionError):
    """Raised when a command fails to execute."""



"""Common internal Accretion utilities."""
from dataclasses import dataclass
from typing import Dict

__all__ = ("PackageVersion",)


@dataclass
class PackageVersion:
    """Container for information identifying a package.

    :param str name: Package name
    :param str version: Package version
    """

    name: str
    version: str

    @classmethod
    def from_pip_log(cls, log_value: str) -> "PackageVersion":
        """Read package name and version from a ``pip install`` log."""
        name, version = log_value.rsplit("-", 1)
        return PackageVersion(name, version)

    def to_dict(self) -> Dict[str, str]:
        """Pack information into a dictionary."""
        return dict(Name=self.name, Version=self.version)



"""Upload a built artifact to S3."""
import hashlib
import os
from typing import Dict, Iterable, Optional

from botocore.client import BaseClient
from botocore.exceptions import ClientError

from accretion_common.constants import ARTIFACTS_PREFIX

from .zipper import build_zip

__all__ = ("artifact_exists", "efficient_build_and_upload_zip")


def _key_hash(installed: Iterable[Dict[str, str]], runtime_name: str, force_new: bool) -> str:
    """Construct a deterministic ID based on the installed requirements and the runtime.

    :param list installed: List of installed package names and versions.
    :param str runtime_name: Lambda runtimes that this artifact supports.
    :param bool force_new: Should we force a new S3 object creation?
    :return: Hash ID.
    :rtype: str
    """
    hasher = hashlib.sha256()

    hasher.update(b"===INSTALLED===")
    for statement in sorted([package["Name"] + "-" + package["Version"] for package in installed]):
        hasher.update(statement.encode("utf-8"))

    hasher.update(b"===RUNTIMES===")
    hasher.update(runtime_name.encode("utf-8"))

    if force_new:
        hasher.update(os.urandom(32))

    return hasher.hexdigest()


def artifact_exists(s3_client: BaseClient, bucket_name: str, artifact_key: str) -> bool:
    """Determine if the specified S3 object exists.

    :param s3_client: Boto3 client to use for S3 interaction.
    :param bucket_name: S3 bucket name.
    :param artifact_key: S3 object key.
    :return: Statement of whether or not object exists.
    :rtype: bool
    """
    try:
        s3_client.head_object(Bucket=bucket_name, Key=artifact_key)
    except ClientError:
        return False
    else:
        return True


def efficient_build_and_upload_zip(
    s3_client: BaseClient,
    project_name: str,
    installed: Iterable[Dict[str, str]],
    bucket_name: str,
    build_dir: str,
    runtime_name: str,
    force_new: Optional[bool] = False,
) -> str:
    """Construct zip file from built artifacts and upload it to S3.

    .. note::

        By default, the S3 key is based on a deterministic hash of the installed requirements.
        This lets us only upload a new object if it is actually new.
        If you want to override this and create a new object anyway, set ``force_new=True``.

    :param s3_client: Boto3 client to use for S3 interaction.
    :param str project_name: Project name to use in S3 key.
    :param installed: List of all installed packages and versions. Used to calculate the S3 key.
    :param str build_dir: Directory from which to collect built resources.
    :param str bucket_name: S3 bucket to use.
    :param str runtime_name: Lambda runtime that this artifact supports. Used to calculate the S3 key.
    :param bool force_new: Should we force a new S3 object creation? Used to calculate the S3 key.
    :return: S3 key containing zip file.
    :rtype: str
    """
    artifact_id = _key_hash(installed=installed, runtime_name=runtime_name, force_new=force_new)
    key = f"{ARTIFACTS_PREFIX}{project_name}/{artifact_id}.zip"

    if artifact_exists(s3_client, bucket_name, key):
        return key

    zip_buffer = build_zip(build_dir)

    s3_client.put_object(Bucket=bucket_name, Key=key, Body=zip_buffer)
    return key



"""Tools for working with venvs."""
import logging
import os
import shutil
import subprocess  # nosec : bandit B404 is addressed by only executing pre-defined commands
import tempfile
import venv
from typing import Dict, Iterable

from accretion_common.exceptions import ExecutionError
from accretion_common.util import PackageVersion

__all__ = ("build_requirements",)
_LOGGER = logging.getLogger(__name__)


def _execute_command(command: str) -> (str, str):
    _LOGGER.debug("Executing command: %s", command)
    proc = subprocess.run(  # nosec : bandit B602 is addressed by only executing pre-defined commands
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8", shell=True
    )
    _LOGGER.debug("============STDOUT============")
    _LOGGER.debug(proc.stdout)
    _LOGGER.debug("============STDERR============")
    _LOGGER.debug(proc.stderr)
    if proc.stderr:
        raise ExecutionError("Failed to execute command")
    return proc.stdout, proc.stderr


def _execute_in_venv(venv_dir: str, command: str) -> (str, str):
    return _execute_command(f"source {venv_dir}/bin/activate; {command}; deactivate")


def _build_venv(venv_dir: str) -> (str, str):
    shutil.rmtree(venv_dir, ignore_errors=True)
    venv.create(venv_dir, clear=True, with_pip=True)
    return _execute_in_venv(venv_dir=venv_dir, command="pip install --no-cache-dir --upgrade pip")


def _build_requirements_file(requirements_file: str, *libraries: str):
    with open(requirements_file, "w") as requirements:
        for library in libraries:
            requirements.write(f"{library}\n")


def _install_requirements_to_build(build_dir: str, requirements_file: str, log_file: str, venv_dir: str) -> (str, str):
    return _execute_in_venv(
        venv_dir=venv_dir,
        # Do not cache in order to save what little disk space we have in Lambda.
        command="pip install --no-cache-dir --upgrade --ignore-installed --no-compile "
        f"--log {log_file} "
        f"-r {requirements_file} --target {build_dir}",
    )


def _parse_install_log(log_file: str) -> Iterable[Dict[str, str]]:
    trigger = "Successfully installed"
    installed = []
    with open(log_file, "r") as logs:
        for line in logs:
            line = line.split(" ", 1)[-1].strip()
            if line.startswith(trigger):
                line = line[len(trigger) :].strip()
                installed = line.split()
                break
    _LOGGER.debug("Installed to layer artifact: %s", installed)
    return [PackageVersion.from_pip_log(log_entry).to_dict() for log_entry in installed]


def build_requirements(build_dir: str, venv_dir: str, requirements: Iterable[str]) -> Iterable[Dict[str, str]]:
    """Build the requested requirements into the target build directory using a newly created venv.

    :param str build_dir: Path to directory into which to build requirements.
    :param str venv_dir: Path to directory to use for venv.
    :param requirements: List of requirements to install.
    """
    _build_venv(venv_dir)

    _, requirements_file = tempfile.mkstemp()
    _, log_file = tempfile.mkstemp()

    try:
        _build_requirements_file(requirements_file, *requirements)
        _install_requirements_to_build(
            build_dir=build_dir, requirements_file=requirements_file, log_file=log_file, venv_dir=venv_dir
        )
        return _parse_install_log(log_file=log_file)
    finally:
        os.remove(requirements_file)
        os.remove(log_file)
